﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.IO;
using System.IO.IsolatedStorage;
using System.Text;
using System.Windows;
using System.Xml.Serialization;

namespace Petzold.Phone.EBookReader
{
    public class BookInfo : IEquatable<BookInfo>, INotifyPropertyChanged
    {
        // Remember to change this if GenerateParagraphs or BookInfo.Paginate is changed!
        const int ProgramPaginationVersion = 1;

        string displayTitle;

        // This event is fired only for DisplayTitle!!!!!
        public event PropertyChangedEventHandler PropertyChanged;

        public BookInfo()
        {
            this.PaginationVersion = ProgramPaginationVersion;
            this.BlankLinesBetweenChapters = 3;
            this.Paragraphs = new List<ParagraphInfo>();
            this.Bookmarks = new ObservableCollection<Bookmark>();
            this.Annotations = new ObservableCollection<Annotation>();
        }

        public int ID { set; get; }

        public string DisplayTitle 
        {
            set
            {
                if (displayTitle != value)
                {
                    displayTitle = value;

                    if (PropertyChanged != null)
                        PropertyChanged(this, new PropertyChangedEventArgs("DisplayTitle"));
                }
            }
            get
            {
                return displayTitle;
            }
        }

        public int PaginationVersion { set; get; }

        public Size PaginationPageSize { set; get; }

        public string PaginationFontFamily { set; get; }

        public double PaginationFontSize { set; get; }

        public int BlankLinesBetweenChapters { set; get; }

        public int CurrentChapterIndex { set; get; }

        public int CurrentPageIndex { set; get; }

        public bool IsChapterAndPageInvalid { set; get; }           // not invalid so much as not current

        public int CurrentCharacterOffset { set; get; }

        public int TotalCharacterCount { set; get; }

        [XmlIgnore]
        public double Percentage
        {
            get
            {
                return 100.0 * CurrentCharacterOffset / TotalCharacterCount;
            }
        }

        public DateTime Creation { set; get; }

        public DateTime LastRead { set; get; }

        [XmlIgnore]
        public List<ParagraphInfo> Paragraphs { set; get; }

        public ObservableCollection<ChapterInfo> Chapters { set; get; }

        public ObservableCollection<Bookmark> Bookmarks { set; get; }

        public ObservableCollection<Annotation> Annotations { set; get; }

        // This is called once after a book is first downloaded
        public static BookInfo Create(int bookID, string bookContents, string friendlyTitle)
        {
            BookInfo bookInfo = new BookInfo
            {
                ID = bookID,
                DisplayTitle = friendlyTitle,
                Creation = DateTime.Now
            };

            // Save the book contents
            string bookFilename = GetBookFilename(bookID);

            using (IsolatedStorageFile storage = IsolatedStorageFile.GetUserStoreForApplication())
            {
                using (IsolatedStorageFileStream stream = storage.CreateFile(bookFilename))
                {
                    using (StreamWriter writer = new StreamWriter(stream))
                    {
                        writer.Write(bookContents);
                    }
                }
            }

            // Quicken the paragraph generation by not reloading book from iso storage
            using (StringReader reader = new StringReader(bookContents))
            {
                bookInfo.GenerateParagraphs(reader);
            }

            bookInfo.Chapters = new ObservableCollection<ChapterInfo>();
            bookInfo.GenerateChapters();

            return bookInfo;
        }

        static string GetInfoFilename(int bookID)
        {
            return bookID.ToString() + ".info";
        }

        static string GetBookFilename(int bookID)
        {
            return bookID.ToString() + ".txt";
        }

        public static BookInfo Load(int bookID)
        {
            string infoFilename = BookInfo.GetInfoFilename(bookID); 
            BookInfo book = null;

            using (IsolatedStorageFile isoStorage = IsolatedStorageFile.GetUserStoreForApplication())
            {
                if (isoStorage.FileExists(infoFilename))
                {
                    using (IsolatedStorageFileStream stream = isoStorage.OpenFile(infoFilename, FileMode.Open))
                    {
                        XmlSerializer serializer = new XmlSerializer(typeof(BookInfo));
                        book = serializer.Deserialize(stream) as BookInfo;
                    }
                }
            }

            if (book != null)
                book.Load();

            return book;
        }

        public void Load()
        {
            GenerateParagraphs();

            if (this.PaginationVersion != ProgramPaginationVersion)
            {
                // Regenerate chapter information
                this.GenerateChapters();

                // Set the new pagination version
                this.PaginationVersion = ProgramPaginationVersion;

                // Chapter and Page also invalid
                this.IsChapterAndPageInvalid = true;
            }
        }

        public void Save()
        {
            using (IsolatedStorageFile isoStorage = IsolatedStorageFile.GetUserStoreForApplication())
            {
                using (IsolatedStorageFileStream stream = isoStorage.CreateFile(BookInfo.GetInfoFilename(this.ID)))
                {
                    XmlSerializer serializer = new XmlSerializer(typeof(BookInfo));
                    serializer.Serialize(stream, this);
                }
            }
        }

        public void Delete()
        {
            using (IsolatedStorageFile isoStorage = IsolatedStorageFile.GetUserStoreForApplication())
            {
                string infoFilename = GetInfoFilename(this.ID);

                if (isoStorage.FileExists(infoFilename))
                    isoStorage.DeleteFile(infoFilename);

                string bookFilename = GetBookFilename(this.ID);

                if (isoStorage.FileExists(bookFilename))
                    isoStorage.DeleteFile(bookFilename);
            }
        }

        public bool Equals(BookInfo that)
        {
            return this.ID == that.ID;
        }

        void GenerateParagraphs()
        {
            string filename = BookInfo.GetBookFilename(this.ID);

            using (IsolatedStorageFile isoStorage = IsolatedStorageFile.GetUserStoreForApplication())
            {
                using (IsolatedStorageFileStream stream = isoStorage.OpenFile(filename, FileMode.Open))
                {
                    using (StreamReader reader = new StreamReader(stream))
                    {
                        GenerateParagraphs(reader);
                    }
                }
            }
        }

        void GenerateParagraphs(TextReader reader)
        {
            int accumulatedCharacterCount = 0;

            string line = null;
            string prevLine = null;
            int prevIndent = 0;
            StringBuilder paragraphBuilder = new StringBuilder();

            while (null != (line = reader.ReadLine()))
            {
                // Trim line of all space and find the line indent
                line = line.TrimEnd(' ');
                int indent = line.Length;
                line = line.TrimStart(' ');
                indent -= line.Length;

                // A new paragraph is always implied with a line of zero length
                bool newParagraph = line.Length == 0;

                if (!newParagraph && paragraphBuilder.Length > 0 && prevLine != null)
                {
                    // This is the case for a list ("contents" for example)
                    //  where each line has a different indent
                    if (indent != prevIndent)
                    {
                        newParagraph = true;
                    }

                    // But some indented text needs to be concatenated (epistoles, for example)
                    else if (indent != 0)
                    {
                        string firstWord = line;
                        int spaceIndex = line.IndexOf(' ');

                        if (spaceIndex != -1)
                            firstWord = firstWord.Substring(0, spaceIndex);

                        // This works for some books, but not for "Phineas Finn" where
                        //  at least the first letter has shorter lines than normal
                        // if (indent + prevLine.Length + 1 + firstWord.Length <= 70)
                        //     newParagraph = true;
                    }
                }

                // Check for a possible paragraph termination
                if (newParagraph)
                {
                    if (paragraphBuilder.Length > 0)
                    {
                        SubstituteCharacters(paragraphBuilder);

                        // Create a new ParagraphInfo for the paragraph
                        ParagraphInfo paragraph = new ParagraphInfo
                        {
                            Indent = prevIndent,
                            Text = paragraphBuilder.ToString(),
                            AccumulatedCharacterCount = accumulatedCharacterCount
                        };
                        this.Paragraphs.Add(paragraph);
                        accumulatedCharacterCount += paragraph.Text.Length;

                        // Clear the StringBuilder
                        paragraphBuilder.Remove(0, paragraphBuilder.Length);
                    }

                    if (line.Length > 0)
                    {
                        // Append the line to the paragraph
                        paragraphBuilder.Append(line);
                    }
                    else
                    {
                        // Add the current line that terminates the paragraph
                        ParagraphInfo emptyParagraph = new ParagraphInfo
                        {
                            Indent = 0,
                            Text = line
                        };
                        this.Paragraphs.Add(emptyParagraph);
                    }
                }
                else
                {
                    // Insert a space between appended lines assembled into paragraph
                    if (paragraphBuilder.Length > 0)
                        paragraphBuilder.Append(" ");

                    paragraphBuilder.Append(line);
                }
                prevLine = line;
                prevIndent = indent;
            }

            // Add the last paragraph if there's one in progress
            if (paragraphBuilder.Length > 0)
            {
                SubstituteCharacters(paragraphBuilder);

                ParagraphInfo paragraph = new ParagraphInfo
                {
                    Indent = prevIndent,
                    Text = paragraphBuilder.ToString(),
                    AccumulatedCharacterCount = accumulatedCharacterCount
                };
                this.Paragraphs.Add(paragraph);
                accumulatedCharacterCount += paragraph.Text.Length;
            }
        }

        // Does fancy quotes and em-dashes
        void SubstituteCharacters(StringBuilder paragraphBuilder)
        {
            // Replace boring double quotes with fancy quotes
            bool isCloseQuote = false;

            for (int i = 0; i < paragraphBuilder.Length; i++)
            {
                if (paragraphBuilder[i] == '"')
                {
                    paragraphBuilder[i] = isCloseQuote ? '\x201D' : '\x201C';
                    isCloseQuote ^= true;
                }
            }

            // Replace double dashes with em dashes separated by spaces
            //  (but only when they're surrounded by non-dashes)
            int index = 0;

            while (index < paragraphBuilder.Length - 1)
            {
                if (paragraphBuilder[index] == '-' &&                       // current is dash
                    paragraphBuilder[index + 1] == '-' &&                   // next one is dash
                    (index == 0 || paragraphBuilder[index - 1] != '-') &&   // there isn't an earlier dash
                    (index == paragraphBuilder.Length - 2 || paragraphBuilder[index + 2] != '-')) // there isn't a later dash
                {
                    if (index == 0 && paragraphBuilder.Length == 2)     // double-dash appears alone
                    {
                        paragraphBuilder.Remove(index, 2);
                        paragraphBuilder.Insert(index, "\x2014");
                    }
                    else if (index == 0)                                // double-dash at beginnning of paragraph
                    {
                        paragraphBuilder.Remove(index, 2);
                        paragraphBuilder.Insert(index, "\x2014 ");
                        index += 1;
                    }
                    else if (index >= paragraphBuilder.Length - 3)      // double-dash not followed by at least two characters
                    {
                        paragraphBuilder.Remove(index, 2);
                        paragraphBuilder.Insert(index, " \x2014");
                        index += 1;
                    }
                    else
                    {
                        // Double quote-followed by em-dash ("Bleak House")
                        if (index > 0 && paragraphBuilder[index - 1] == '\x201C')
                        {
                            paragraphBuilder.Remove(index, 2);
                            paragraphBuilder.Insert(index, "\x2014 ");
                            index += 1;
                        }
                        // em-dash followed by double-quote ("Ditto")
                        else if (index < paragraphBuilder.Length - 1 && paragraphBuilder[index + 2] == '\x201D')
                        {
                            paragraphBuilder.Remove(index, 2);
                            paragraphBuilder.Insert(index, " \x2014");
                            index += 1;
                        }
                        else
                        {
                            paragraphBuilder.Remove(index, 2);
                            paragraphBuilder.Insert(index, " \x2014 ");
                            index += 2;
                        }
                    }
                }

                index += 1;
            }
        }

        // This method is called when a BookInfo is first created from the Create
        //  method, but it can also be called later on from the 
        //  ChapterAnalysisPage when the user wishes to redefine the chapters list.
        public void GenerateChapters()
        {
            // Clear the chapter list
            this.Chapters.Clear();

            // Set so the beginning of the book becomes the first chapter
            int consecutiveBlankLineCount = 10;
            int accumulatedCharacterCount = 0;
            ChapterInfo chapterInfo = null;

            for (int paragraphIndex = 0; paragraphIndex < this.Paragraphs.Count; paragraphIndex++)
            {
                if (this.Paragraphs[paragraphIndex].Text.Length == 0)
                {
                    consecutiveBlankLineCount += 1;
                }
                else
                {
                    if (consecutiveBlankLineCount >= this.BlankLinesBetweenChapters)
                    {
                        PageInfo pageInfo = new PageInfo
                        {
                            ParagraphIndex = paragraphIndex,
                            CharacterIndex = 0,
                            AccumulatedCharacterCount = accumulatedCharacterCount
                        };

                        chapterInfo = new ChapterInfo
                        {
                            Title = this.Paragraphs[paragraphIndex].Text.Length < 100 ?
                                        this.Paragraphs[paragraphIndex].Text :
                                        this.Paragraphs[paragraphIndex].Text.Substring(0, 100),
                            FirstParagraph = paragraphIndex,
                            ParagraphCount = 1,
                            IsFirstChapter = this.Chapters.Count == 0
                        };
                        chapterInfo.Pages.Add(pageInfo);
                        this.Chapters.Add(chapterInfo);
                    }
                    else
                    {
                        // Accumulate number of paragraphs in the chapter
                        chapterInfo.ParagraphCount += 1 + consecutiveBlankLineCount;
                    }
                    consecutiveBlankLineCount = 0;
                    accumulatedCharacterCount += this.Paragraphs[paragraphIndex].Text.Length;
                }
            }

            // Flag the last chapter
            this.Chapters[this.Chapters.Count - 1].IsLastChapter = true;
            this.TotalCharacterCount = accumulatedCharacterCount;
        }

        // Needs to be called whenever the chapter definitions change.
        public void UpdateBookmarksAndAnnotations()
        {
            foreach (Bookmark bookmark in this.Bookmarks)
                bookmark.ChapterTitle = GetChapterTitle(bookmark.CharacterOffset);

            foreach (Annotation annotation in this.Annotations)
                annotation.ChapterTitle = GetChapterTitle(annotation.CharacterOffset);
        }

        string GetChapterTitle(int characterOffset)
        {
            string chapterTitle = null;

            foreach (ChapterInfo chapter in this.Chapters)
            {
                if (characterOffset < chapter.Pages[0].AccumulatedCharacterCount)
                    break;

                chapterTitle = chapter.Title;
            }
            return chapterTitle;
        }
    }
}
