﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Xml.Serialization;

namespace Petzold.Phone.EBookReader.RdfDeserialization
{
    // Project Gutenberg maintains a file http://www.gutenberg.org/ebooks/{id}.rdf for each book.
    // I had not planned to use these files, since it seemed I could get sufficient stuff from
    //      the large catalog file, and I needed that anyway.
    // However, for some books, e.g., 642, the UTF-8 file indicated in the catalog was empty,
    //      and a different filename appeared in the RDF file.
    // Moreover, for other books, e.g., 

    class XmlNamespace
    {
        public const string cc = "http://web.resource.org/cc/";
        public const string dcterms = "http://purl.org/dc/terms/";
        public const string rdf = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
        public const string pgterms = "http://www.gutenberg.org/2009/pgterms/";
    }

    [XmlRoot(Namespace = XmlNamespace.rdf)]
    public class RDF
    {
        // Valid for catalog and individual books
        [XmlElement(ElementName = "Description", Namespace = XmlNamespace.rdf)]
        public Description Description { set; get; }

        // Valid for catalog only
        [XmlElement(ElementName = "etext", Namespace = XmlNamespace.pgterms)]
        public List<etext> etext { set; get; }

        // Valid for individual books
        [XmlElement(ElementName = "Work", Namespace = XmlNamespace.cc)]
        public Work Work { set; get; }

        [XmlElement(ElementName = "ebook", Namespace = XmlNamespace.pgterms)]
        public ebook ebook { set; get; }

        [XmlElement(ElementName = "agent", Namespace = XmlNamespace.pgterms)]
        public List<agent> agents { set; get; }

        [XmlElement(ElementName = "file", Namespace = XmlNamespace.pgterms)]
        public List<file> Files { set; get; }

        public static RDF Deserialize(string xml)
        {
            using (StringReader reader = new StringReader(xml))
            {
                return RDF.Deserialize(reader);
            }
        }

        public static RDF Deserialize(TextReader reader)
        {
            XmlSerializer serializer = new XmlSerializer(typeof(RDF));
            return serializer.Deserialize(reader) as RDF;
        }
    }

    public class Description
    {
        [XmlElement(ElementName = "value", Namespace = XmlNamespace.rdf)]
        public List<value> Values { set; get; }
    }

    public class etext
    {
        [XmlElement(ElementName = "publisher", Namespace = XmlNamespace.dcterms)]
        public string publisher { set; get; }

        [XmlElement(ElementName = "title", Namespace = XmlNamespace.dcterms)]
        public string title { set; get; }

        [XmlElement(ElementName = "creator", Namespace = XmlNamespace.dcterms)]
        public List<string> creator { set; get; }

        [XmlElement(ElementName = "friendlytitle", Namespace = XmlNamespace.dcterms)]
        public string friendlytitle { set; get; }
    }

    public class Work
    {

    }

    public class ebook
    {
        [XmlElement(ElementName = "title", Namespace = XmlNamespace.dcterms)]
        public string title { set; get; }

        [XmlElement(ElementName = "publisher", Namespace = XmlNamespace.dcterms)]
        public string publisher { set; get; }

        [XmlElement(ElementName = "issued", Namespace = XmlNamespace.dcterms)]
        public string issued { set; get; }

        [XmlElement(ElementName = "language", Namespace = XmlNamespace.dcterms)]
        public string language { set; get; }

        [XmlElement(ElementName = "type", Namespace = XmlNamespace.dcterms)]
        public fileType type { set; get; }
    }

    public class fileType
    {
        [XmlElement(ElementName = "Description", Namespace = XmlNamespace.rdf)]
        public Description Description { set; get; }
    }

    public class agent
    {
        [XmlElement(ElementName = "name", Namespace = XmlNamespace.pgterms)]
        public string name { set; get; }

        [XmlElement(ElementName = "birthdate", Namespace = XmlNamespace.pgterms)]
        public string birthdate { set; get; }

        [XmlElement(ElementName = "deathdate", Namespace = XmlNamespace.pgterms)]
        public string deathdate { set; get; }

        public override string ToString()
        {
            if (!String.IsNullOrEmpty(birthdate))
            {
                if (!String.IsNullOrEmpty(deathdate))
                    return String.Format("{0} ({1} - {2})", name, birthdate, deathdate);
                else
                    return String.Format("{0} (b. {1})", name, birthdate);
            }

            return name;
        }
    }

    public class value
    {
        [XmlText]
        public string content { set; get; }
    }

    public class file
    {
        [XmlAttribute(Namespace = XmlNamespace.pgterms)]
        public string about { set; get; }

        [XmlElement(ElementName = "extent", Namespace = XmlNamespace.dcterms)]
        public string extent { set; get; }

        [XmlElement(ElementName = "format", Namespace = XmlNamespace.dcterms)]
        public format format { set; get; }
    }

    public class format
    {
        [XmlElement(ElementName = "Description", Namespace = XmlNamespace.rdf)]
        public Description Description { set; get; }
    }
}
