﻿using System;
using System.Collections.Generic;
using System.IO;
using System.IO.IsolatedStorage;
using System.Windows;
using System.Windows.Navigation;
using System.Xml.Serialization;
using Microsoft.Phone.Controls;
using Microsoft.Phone.Shell;
using Petzold.Phone.EBookReader;

namespace PhreeBookReader
{
    public partial class App : Application
    {
        // Field for saving and restoring app settings
        const string appSettingsFilename = "PhreeBookReader.settings";

        // Properties for saving and restoring app & book settings
        public AppSettings AppSettings { private set; get; }
        public BookInfo CurrentBook { set; get; }

        // Property for miscelleneous settings
        public IDictionary<string, object> TombstoneObjects { set; get; }

        /// <summary>
        /// Provides easy access to the root frame of the Phone Application.
        /// </summary>
        /// <returns>The root frame of the Phone Application.</returns>
        public PhoneApplicationFrame RootFrame { get; private set; }

        /// <summary>
        /// Constructor for the Application object.
        /// </summary>
        public App()
        {
            // Global handler for uncaught exceptions. 
            UnhandledException += Application_UnhandledException;

            // Show graphics profiling information while debugging.
            if (System.Diagnostics.Debugger.IsAttached)
            {
                // Display the current frame rate counters.
                Application.Current.Host.Settings.EnableFrameRateCounter = true;

                // Show the areas of the app that are being redrawn in each frame.
                //Application.Current.Host.Settings.EnableRedrawRegions = true;

                // Enable non-production analysis visualization mode, 
                // which shows areas of a page that are being GPU accelerated with a colored overlay.
                //Application.Current.Host.Settings.EnableCacheVisualization = true;
            }

            // Standard Silverlight initialization
            InitializeComponent();

            // Phone-specific initialization
            InitializePhoneApplication();

            // Dump Isolated Storage is the debugger is attached
            if (System.Diagnostics.Debugger.IsAttached)
            {
                using (IsolatedStorageFile isoStorage = IsolatedStorageFile.GetUserStoreForApplication())
                {
                    DumpDirectories(isoStorage, "*.*", 0);
                }
            }
        }

        void DumpDirectories(IsolatedStorageFile isoStorage, string pattern, int indent)
        {
            DumpFiles(isoStorage, pattern, indent);

            string[] directoryNames = null;

            try
            {
                directoryNames = isoStorage.GetDirectoryNames(pattern);
            }
            catch
            {
                return;
            }

            Array.Sort(directoryNames);

            foreach (string directoryName in directoryNames)
            {
                System.Diagnostics.Debug.WriteLine(String.Format("{0}{1}", new string(' ', 4 * indent), directoryName));
                DumpDirectories(isoStorage, Path.Combine(directoryName, pattern), indent + 1);
            }
        }

        void DumpFiles(IsolatedStorageFile isoStorage, string pattern, int indent)
        {
            string[] filenames = null;

            try
            {
                filenames = isoStorage.GetFileNames(pattern);
            }
            catch
            {
                return;
            }

            Array.Sort(filenames);

            foreach (string filename in filenames)
            {
                using (IsolatedStorageFileStream stream = isoStorage.OpenFile(filename, FileMode.Open, FileAccess.Read, FileShare.Read))
                {
                    System.Diagnostics.Debug.WriteLine("{0}{1} - {2} bytes", new string(' ', 4 * indent), filename, stream.Length);
                }
            }
        }

        // Code to execute when the application is launching (eg, from Start)
        // This code will not execute when the application is reactivated
        private void Application_Launching(object sender, LaunchingEventArgs e)
        {
            LoadSettings(false);
        }

        // Code to execute when the application is activated (brought to foreground)
        // This code will not execute when the application is first launched
        private void Application_Activated(object sender, ActivatedEventArgs e)
        {
            LoadSettings(true);
        }

        // Code to execute when the application is deactivated (sent to background)
        // This code will not execute when the application is closing
        private void Application_Deactivated(object sender, DeactivatedEventArgs e)
        {
            SaveSettings(true);
        }

        // Code to execute when the application is closing (eg, user hit Back)
        // This code will not execute when the application is deactivated
        private void Application_Closing(object sender, ClosingEventArgs e)
        {
            SaveSettings(false);
        }

        void LoadSettings(bool tombstoning)
        {
            // Application Settings from isolated storage
            this.AppSettings = AppSettings.Load(appSettingsFilename);

            // Miscellaneous tombstone objects
            this.TombstoneObjects = new Dictionary<string, object>();

            if (tombstoning)
            {
                PhoneApplicationService appService = PhoneApplicationService.Current;

                // Current book
                if (appService.State.ContainsKey("currentBookID"))
                {
                    int bookID = (int)appService.State["currentBookID"];
                    appService.State.Remove("currentBookID");
                    this.CurrentBook = BookInfo.Load(bookID);
                }

                // Transfer miscellaneous tombstone objects from ApplicationService state dictionary
                foreach (string key in appService.State.Keys)
                    if (key.StartsWith("tombstoneObjects"))
                        this.TombstoneObjects.Add(key.Substring(16), appService.State[key]);

                // Then remove them from ApplicationService state dictionary
                foreach (string key in this.TombstoneObjects.Keys)
                    appService.State.Remove("tombstoneObjects" + key);
            }
        }

        void SaveSettings(bool tombstoning)
        {
            // Save application settings
            this.AppSettings.Save(appSettingsFilename);

            if (tombstoning)
            {
                PhoneApplicationService appService = PhoneApplicationService.Current;

                // Save current book information
                if (this.CurrentBook != null)
                {
                    appService.State.Add("currentBookID", this.CurrentBook.ID);
                    this.CurrentBook.Save();
                }

                // Save miscellaneous tombstone objects
                if (this.TombstoneObjects.Count > 0)
                {
                    foreach (string key in this.TombstoneObjects.Keys)
                        appService.State.Add("tombstoneObjects" + key, this.TombstoneObjects[key]);
                }
            }
        }

        // Code to execute if a navigation fails
        private void RootFrame_NavigationFailed(object sender, NavigationFailedEventArgs e)
        {
            if (System.Diagnostics.Debugger.IsAttached)
            {
                // A navigation has failed; break into the debugger
                System.Diagnostics.Debugger.Break();
            }
        }

        // Code to execute on Unhandled Exceptions
        private void Application_UnhandledException(object sender, ApplicationUnhandledExceptionEventArgs e)
        {
            if (System.Diagnostics.Debugger.IsAttached)
            {
                // An unhandled exception has occurred; break into the debugger
                System.Diagnostics.Debugger.Break();
            }
        }

        #region Phone application initialization

        // Avoid double-initialization
        private bool phoneApplicationInitialized = false;

        // Do not add any additional code to this method
        private void InitializePhoneApplication()
        {
            if (phoneApplicationInitialized)
                return;

            // Create the frame but don't set it as RootVisual yet; this allows the splash
            // screen to remain active until the application is ready to render.
            RootFrame = new PhoneApplicationFrame();
            RootFrame.Navigated += CompleteInitializePhoneApplication;

            // Handle navigation failures
            RootFrame.NavigationFailed += RootFrame_NavigationFailed;

            // Ensure we don't initialize again
            phoneApplicationInitialized = true;
        }

        // Do not add any additional code to this method
        private void CompleteInitializePhoneApplication(object sender, NavigationEventArgs e)
        {
            // Set the root visual to allow the application to render
            if (RootVisual != RootFrame)
                RootVisual = RootFrame;

            // Remove this handler since it is no longer needed
            RootFrame.Navigated -= CompleteInitializePhoneApplication;
        }

        #endregion
    }
}