﻿using System;
using System.IO.IsolatedStorage;
using System.Net;
using System.Windows;
using System.Windows.Navigation;
using Microsoft.Phone.Controls;
using Microsoft.Phone.Shell;
using Petzold.Phone.EBookReader;
using Petzold.Phone.EBookReader.RdfDeserialization;
using PhreeBookReader.CatalogServiceReference;

namespace PhreeBookReader
{
    public partial class DownloadBookPage : PhoneApplicationPage
    {
        PhoneApplicationService appService = PhoneApplicationService.Current;
        AppSettings appSettings = (Application.Current as App).AppSettings;

        GutenbergBook gutenbergBook;
        BookInfo bookInfo;
        WebClient webClient = new WebClient();
        
        enum DownloadMode { RdfFile, BookFile };
        DownloadMode downloadMode;

        public DownloadBookPage()
        {
            InitializeComponent();

            webClient.DownloadProgressChanged += OnBookFileDownloadProgressChanged;
            webClient.DownloadStringCompleted += OnBookFileDownloadStringCompleted;

            Loaded += new RoutedEventHandler(OnLoaded);
        }

        protected override void OnNavigatedTo(NavigationEventArgs args)
        {
            // Case 1:
            //      Navigating to this page from "search" pivot or "request" pivot on MainPage,
            //      Or: returning from tombstoning without having yet downloaded the book.

            if (appService.State.ContainsKey("downloadBook"))
            {
                gutenbergBook = appService.State["downloadBook"] as GutenbergBook;
                catalogInfo.GutenbergBook = gutenbergBook;                

                // Test is enough space is available for download
                long freeSpace = 0;

                using (IsolatedStorageFile storage = IsolatedStorageFile.GetUserStoreForApplication())
                {
                    freeSpace = storage.AvailableFreeSpace;
                }

                if (1.5 * gutenbergBook.FileSize > freeSpace)
                {
                    requiredStorageText.Text = (1.5 * gutenbergBook.FileSize / 1024).ToString("N0");
                    availableStorageText.Text = (freeSpace / 1024).ToString("N0");
                    insufficientSpaceTextBlock.Visibility = Visibility.Visible;
                    downloadButton.IsEnabled = false;
                }
                else
                {
                    // If space is OK, set a warning about a duplicate book
                    duplicateWarningTextBlock.Visibility = appSettings.Library.DoesLibraryEntryExist(gutenbergBook.ID) ? Visibility.Visible : Visibility.Collapsed;
                }
            }

            // Case 2:
            //      Navigating back to this page from DisplayTitlePage or ChapterAnalysisPage,
            //      Or: returning from tombstoning after having downloaded the book.

            else if (appService.State.ContainsKey("successfulDownload"))
            {
                gutenbergBook = appService.State["successfulDownload"] as GutenbergBook;
                catalogInfo.GutenbergBook = gutenbergBook;

                bookInfo = BookInfo.Load(gutenbergBook.ID);
                bookTitleTextBlock.Text = bookInfo.DisplayTitle;

                downloadButton.IsEnabled = false;
                downloadTextBlock.Text = "Download successful!";
                downloadTextBlock.TextAlignment = TextAlignment.Center;
                optionsPanel.Visibility = Visibility.Visible;
            }

            // Case 3:
            //      Returning to this page after reading the newly-downloaded book

            else // if (appService.State.ContainsKey("successfulDownloadAndRead"))
            {
                this.NavigationService.GoBack();
            }

            base.OnNavigatedTo(args);
        }

        protected override void OnNavigatedFrom(NavigationEventArgs e)
        {
            base.OnNavigatedFrom(e);
        }

        void OnLoaded(object sender, RoutedEventArgs args)
        {
            // Scroll to where the buttons are
            scrollViewer.ScrollToBottom();
        }

        void OnDownloadButtonClick(object sender, RoutedEventArgs args)
        {
            // Switch to "cancel" button
            downloadButton.IsEnabled = false;
            downloadButton.Visibility = Visibility.Collapsed;
            cancelButton.IsEnabled = true;
            cancelButton.Visibility = Visibility.Visible;

            // Initialize and start download of RDF file
            downloadTextBlock.Text = null;
            progressBar.Value = 0;
            downloadMode = DownloadMode.RdfFile;
            string rdfFilename = String.Format("http://www.gutenberg.org/ebooks/{0}.rdf", gutenbergBook.ID);
            webClient.DownloadStringAsync(new Uri(rdfFilename));
        }

        void OnCancelButtonClick(object sender, RoutedEventArgs args)
        {
            // Disable the button
            cancelButton.IsEnabled = false;

            // Cancel the download
            webClient.CancelAsync();
        }

        void OnBookFileDownloadProgressChanged(object sender, DownloadProgressChangedEventArgs args)
        {
            // Check if OK to access UI
            if (this.NavigationService.CanGoBack && downloadMode == DownloadMode.BookFile)
            {
                progressBar.Value = args.ProgressPercentage;
            }
        }

        void OnBookFileDownloadStringCompleted(object sender, DownloadStringCompletedEventArgs args)
        {
            if (args.Cancelled)
            {
                DownloadCompleted("Download cancelled.", null, false, true);
            }
            else if (args.Error != null)
            {
                DownloadCompleted(args.Error.Message, null, false, true);
            }
            else if (String.IsNullOrEmpty(args.Result))
            {
                DownloadCompleted("Downloaded book has no content.", null, false, false);
            }
            else
            {
                if (downloadMode == DownloadMode.BookFile)
                {
                    DownloadCompleted("Download successful!", args.Result, true, false);
                }
                else // downloadMode == DownloadMode.RdfFile
                {
                    // Replace the filename with a UTF-8 file from the RDF
                    RDF rdf = RDF.Deserialize(args.Result);
                    string filename;

                    if (GetReplacementFile(rdf, "text/plain; charset=utf-8", out filename))
                    {
                        gutenbergBook.Format = "text/plain; charset=utf-8";
                        gutenbergBook.Filename = filename;
                    }
                    else if (GetReplacementFile(rdf, "text/plain", out filename))
                    {
                        gutenbergBook.Format = "text/plain";
                        gutenbergBook.Filename = filename;
                    }

                    // Regardless whether a replacement UTF-8 was found or not, download the file
                    downloadMode = DownloadMode.BookFile;
                    webClient.DownloadStringAsync(new Uri(gutenbergBook.Filename));
                }
            }
        }

        bool GetReplacementFile(RDF rdf, string format, out string filename)
        {
            foreach (file file in rdf.Files)
            {
                if (file.format.Description.Values.Count == 1 &&
                    file.format.Description.Values[0].content == format)
                {
                    filename = file.about;
                    return true;
                }
            }
            filename = null;
            return false;
        }

        void DownloadCompleted(string message, string bookContents, bool isSuccess, bool canTryAgain)
        {
            // Check if the page is still active before accessing controls
            if (this.NavigationService.CanGoBack)
            {
                downloadTextBlock.Text = message;

                if (isSuccess)
                    downloadTextBlock.TextAlignment = TextAlignment.Center;

                downloadButton.Visibility = Visibility.Visible;
                cancelButton.Visibility = Visibility.Collapsed;
                downloadButton.IsEnabled = canTryAgain;

                scrollViewer.ScrollToBottom();
            }

            if (isSuccess)
            {
                // Save the GutenbergBook object in isolated storage using an extension method.
                gutenbergBook.Save();

                // Create a new BookInfo (which also saves the bookContents).
                bookInfo = BookInfo.Create(gutenbergBook.ID, bookContents, gutenbergBook.FriendlyTitle);

                // Save the new BookInfo.
                bookInfo.Save();

                // Check if the book is already in the library; if so, remove it.
                if (appSettings.Library.DoesLibraryEntryExist(gutenbergBook.ID) && appSettings.Bedside.Contains(bookInfo))
                    appSettings.Bedside.Remove(bookInfo);

                // Add the new book to the library and save.
                appSettings.Library.CreateLibraryEntry(gutenbergBook.ID, gutenbergBook.Titles, gutenbergBook.Creators);
                
                // Set a global application state indicating the file has been successfully downloaded.
                StateDictionaryHelper.SafeRemove(appService.State, "downloadBook");
                StateDictionaryHelper.SafeAdd(appService.State, "successfulDownload", gutenbergBook);

                // Check if this page is still active before accessing controls
                if (this.NavigationService.CanGoBack)
                {
                    duplicateWarningTextBlock.Visibility = Visibility.Collapsed;
                    optionsPanel.Visibility = Visibility.Visible;
                    scrollViewer.ScrollToBottom();
                }
            }
        }

        // These two pages load and save the BookInfo
        void OnChangeDisplayTitleClick(object sender, RoutedEventArgs args)
        {
            string uri = String.Format("/DisplayTitlePage.xaml?bookid={0}", bookInfo.ID);
            this.NavigationService.Navigate(new Uri(uri, UriKind.Relative));
        }

        void OnChapterBreaksClick(object sender, RoutedEventArgs args)
        {
            string uri = String.Format("/ChapterAnalysisPage.xaml?bookid={0}", bookInfo.ID);
            this.NavigationService.Navigate(new Uri(uri, UriKind.Relative));
        }

        void OnDoneButtonClick(object sender, RoutedEventArgs args)
        {
            this.NavigationService.GoBack();
        }

        void OnReadButtonClick(object sender, RoutedEventArgs args)
        {
            StateDictionaryHelper.SafeRemove(appService.State, "successfulDownload");
            StateDictionaryHelper.SafeAdd(appService.State, "successfulDownloadAndRead", gutenbergBook.ID);

            (Application.Current as App).CurrentBook = bookInfo;
            this.NavigationService.Navigate(new Uri("/BookViewerPage.xaml", UriKind.Relative));
        }
    }
}