﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Navigation;
using Microsoft.Phone.Controls;
using Microsoft.Phone.Shell;
using Petzold.Phone.EBookReader;

namespace PhreeBookReader
{
    public partial class MainPage : PhoneApplicationPage
    {
        PhoneApplicationService appService = PhoneApplicationService.Current;
        AppSettings appSettings = (Application.Current as App).AppSettings;
        int lastPivotSelectedIndex = -1;
        double libraryScrollOffset = 0;

        public MainPage()
        {
            InitializeComponent();
            besideListBox.ItemsSource = appSettings.Bedside;
            libraryItemsControl.ItemsSource = appSettings.Library.Authors;
        }

        protected override void OnNavigatedFrom(NavigationEventArgs args)
        {
            StateDictionaryHelper.SafeAdd(this.State, "lastPivotSelectedIndex", pivot.SelectedIndex);
            StateDictionaryHelper.SafeAdd(this.State, "libraryScrollOffset", libraryScrollViewer.VerticalOffset);

            foreach (PivotItem pivotItem in pivot.Items)
            {
                if (pivotItem.Content is ITombstonable)
                    (pivotItem.Content as ITombstonable).SaveState(pivotItem.Header.ToString(), this.State);
            }

            base.OnNavigatedFrom(args);
        }

        protected override void OnNavigatedTo(NavigationEventArgs args)
        {
            lastPivotSelectedIndex = StateDictionaryHelper.SafeGet(this.State, "lastPivotSelectedIndex", -1);
            libraryScrollOffset = StateDictionaryHelper.SafeGet(this.State, "libraryScrollOffset", 0.0);

            BookInfo book = (Application.Current as App).CurrentBook;

            // If we're navigating back here, and App.CurrentBook is != null,
            //  then we're done reading that book for now, and the BookInfo 
            //  file needs to be saved.
            if (book != null)
            {
                book.Save();
                (Application.Current as App).CurrentBook = null;
            }

            foreach (PivotItem pivotItem in pivot.Items)
            {
                if (pivotItem.Content is ITombstonable)
                    (pivotItem.Content as ITombstonable).RestoreState(pivotItem.Header.ToString(), this.State);
            }

            base.OnNavigatedTo(args);
        }

        void OnPivotLoaded(object sender, RoutedEventArgs args)
        {
            // Program starting up
            if (lastPivotSelectedIndex == -1)
            {
                // Go to bedside if books are there
                if (appSettings.Bedside.Count != 0)
                {
                    pivot.SelectedIndex = 0;
                }
                // Or library if books are there
                else if (appSettings.Library.Authors.Count != 0)
                {
                    pivot.SelectedIndex = 1;
                }
                // Otherwise go "shopping"
                else
                {
                    pivot.SelectedIndex = 2;
                }
            }
            else if (lastPivotSelectedIndex == 0 && appSettings.Bedside.Count != 0)
            {
                pivot.SelectedIndex = 0;
            }
            else if (lastPivotSelectedIndex <= 1 && appSettings.Library.Authors.Count != 0)
            {
                pivot.SelectedIndex = 1;
            }
            else if (lastPivotSelectedIndex == 4)   // about
            {
                pivot.SelectedIndex = 4;
            }
            else if (lastPivotSelectedIndex == 2 || lastPivotSelectedIndex == 3)    // download portals
            {
                // Coming back from successful download after user has pressed "read book"
                if (appService.State.ContainsKey("successfulDownloadAndRead"))
                {
                    appService.State.Remove("successfulDownloadAndRead");
                    pivot.SelectedIndex = 0;    // bedside
                }
                // Coming back from successful download 
                else if (appService.State.ContainsKey("successfulDownload"))
                {
                    appService.State.Remove("successfulDownload");
                    pivot.SelectedIndex = 1;    // library
                }
                // Coming back from DownloadBookPage without downloading
                else
                {
                     if (appService.State.ContainsKey("downloadFile"))
                         appService.State.Remove("downloadFile");

                     pivot.SelectedIndex = lastPivotSelectedIndex;
                }
            }

            // Wonder why this isn't working?
            libraryScrollViewer.ScrollToVerticalOffset(libraryScrollOffset);
            libraryScrollViewer.UpdateLayout();
        }

        void OnInfoButtonClick(object sender, RoutedEventArgs args)
        {
            Button btn = sender as Button;
            int bookID = (int)btn.Tag;
            string uri = String.Format("/BookInformationPage.xaml?bookid={0}", bookID);
            this.NavigationService.Navigate(new Uri(uri, UriKind.Relative));
        }

        void OnBedsideListBoxSelectionChanged(object sender, SelectionChangedEventArgs args)
        {
            ListBox lstbox = sender as ListBox;
            BookInfo book = lstbox.SelectedItem as BookInfo;
            lstbox.SelectedItem = null;

            if (book != null)
            {
                // Call method to generate paragraphs
                book.Load();
                NavigateToBookViewerPage(book);
            }
        }

        void OnLibraryListBoxSelectionChanged(object sender, SelectionChangedEventArgs args)
        {
            ListBox lstbox = sender as ListBox;
            TitleInfo title = lstbox.SelectedItem as TitleInfo;
            lstbox.SelectedItem = null;

            if (title != null)
            {
                int bookID = title.ID;
                NavigateToBookViewerPage(bookID);
            }
        }

        void NavigateToBookViewerPage(int bookID)
        {
            BookInfo book = BookInfo.Load(bookID);
            NavigateToBookViewerPage(book);
        }

        void NavigateToBookViewerPage(BookInfo book)
        {
            // Store BookInfo object in App.CurrentBook property for reference by all pages
            (Application.Current as App).CurrentBook = book;

            // Navigate to the BookViewerPage
            this.NavigationService.Navigate(new Uri("/BookViewerPage.xaml", UriKind.Relative));
        }
    }
}