﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Threading;
using Microsoft.Phone.Shell;
using Petzold.Phone.EBookReader;
using PhreeBookReader.CatalogServiceReference;

namespace PhreeBookReader
{
    public partial class SearchControl : UserControl, ITombstonable
    {
        CatalogServiceClient proxy = new CatalogServiceClient();
        CatalogStats catalogStats = null;

        // Consolidate these three fields for tombstoning
        public struct SearchParams
        {
            public string Title { set; get; }
            public string Author { set; get; }
            public int Index { set; get; }
            public string SearchKey { set; get; }
        }
        SearchParams searchParams;

        public SearchControl()
        {
            InitializeComponent();
            proxy.GetStatsCompleted += OnCatalogServiceProxyGetStatsCompleted;
            proxy.SearchCompleted += OnCatalogServiceProxySearchCompleted;
            proxy.GetStatsAsync();
        }

        // ITombstonable method, called from MainPage
        public void SaveState(string prefix, IDictionary<string, object> state)
        {
            StateDictionaryHelper.SafeAdd(state, prefix + "titleTextBox", titleTextBox.Text);
            StateDictionaryHelper.SafeAdd(state, prefix + "authorTextBox", authorTextBox.Text);
            StateDictionaryHelper.SafeAdd(state, prefix + "searchParams", searchParams);
            StateDictionaryHelper.SafeAdd(state, prefix + "statusText", statusTextBlock.Text);
            StateDictionaryHelper.SafeAdd(state, prefix + "nextSearchText", nextSearchButton.Content);
            StateDictionaryHelper.SafeAdd(state, prefix + "nextSearchVisibility", nextSearchButton.Visibility);

            if (resultsListBox.Items.Count != 0)
            {
                List<GutenbergBook> items = new List<GutenbergBook>();

                foreach (object item in resultsListBox.Items)
                    items.Add(item as GutenbergBook);

                IDictionary<string, object> tombstoneObjects = (Application.Current as App).TombstoneObjects;
                StateDictionaryHelper.SafeAdd(tombstoneObjects, prefix + "listBoxItems", items);
            }
        }

        // ITombstonable method, called from MainPage
        public void RestoreState(string prefix, IDictionary<string, object> state)
        {
            titleTextBox.Text = StateDictionaryHelper.SafeGet(state, prefix + "titleTextBox", "");
            authorTextBox.Text = StateDictionaryHelper.SafeGet(state, prefix + "authorTextBox", "");
            searchParams = StateDictionaryHelper.SafeGet(state, prefix + "searchParams", new SearchParams());
            statusTextBlock.Text = StateDictionaryHelper.SafeGet<string>(state, prefix + "statusText", null);
            nextSearchButton.Content = StateDictionaryHelper.SafeGet(state, prefix + "nextSearchText", "search");
            nextSearchButton.Visibility = StateDictionaryHelper.SafeGet(state, prefix + "nextSearchVisibility", Visibility.Collapsed);

            // Only restore ListBox contents if we need to
            if (resultsListBox.Items.Count == 0)
            {
                IDictionary<string, object> tombstoneObjects = (Application.Current as App).TombstoneObjects;
                List<GutenbergBook> items = StateDictionaryHelper.SafeGet(tombstoneObjects, prefix + "listBoxItems", new List<GutenbergBook>());

                foreach (GutenbergBook item in items)
                    resultsListBox.Items.Add(item);
            }
        }

        void OnCatalogServiceProxyGetStatsCompleted(object sender, GetStatsCompletedEventArgs args)
        {
            if (args.Cancelled)
            {
                return;
            }
            else if (args.Error != null)
            {
                // Try again in 10 seconds
                DispatcherTimer tmr = new DispatcherTimer { Interval = TimeSpan.FromSeconds(10) };
                tmr.Tick += (tmrSender, tmrArgs) =>
                    {
                        tmr.Stop();
                        proxy.GetStatsAsync();
                    };
                tmr.Start();
            }
            else if (catalogInfoTextBlock != null)
            {
                catalogStats = args.Result;
                catalogInfoTextBlock.Text = String.Format("Catalog dated {0} lists {1} books", 
                                                          catalogStats.CatalogDate, catalogStats.BookCount);
            }
        }

        void OnTextBoxTextChanged(object sender, TextChangedEventArgs args)
        {
            searchButton.IsEnabled = titleTextBox.Text.Length > 0 || authorTextBox.Text.Length > 0;
        }

        protected override void OnKeyUp(KeyEventArgs args)
        {
            if (args.Key == Key.Enter)
            {
                OnSearchButtonClick(this, args);
                args.Handled = true;
            }
            base.OnKeyUp(args);
        }

        void OnSearchButtonClick(object sender, RoutedEventArgs args)
        {
            EnableForSearch(false);
            statusTextBlock.Text = null;
            nextSearchButton.Visibility = Visibility.Collapsed;
            resultsListBox.Items.Clear();

            searchParams.Title = titleTextBox.Text;
            searchParams.Author = authorTextBox.Text;
            searchParams.Index = 0;

            proxy.GetStatsAsync();
            proxy.SearchAsync(searchParams.Title, searchParams.Author, searchParams.Index, null);
        }

        void OnCatalogServiceProxySearchCompleted(object sender, SearchCompletedEventArgs args)
        {
            EnableForSearch(true);

            if (args.Cancelled)
            {
                statusTextBlock.Text = "Search was cancelled";
            }
            else if (args.Error != null)
            {
                statusTextBlock.Text = args.Error.Message;
            }
            else
            {
                SearchResults searchResults = args.Result;

                // Check if catalog stats needs to be updated
                if (catalogStats != null && catalogStats.SearchKey != searchResults.SearchKey)
                {
                    proxy.GetStatsAsync();
                }

                if (searchResults.Books == null || searchResults.Books.Count == 0)
                {
                    statusTextBlock.Text = "No matches found";
                    nextSearchButton.Visibility = Visibility.Collapsed;
                }
                else
                {
                    foreach (GutenbergBook book in searchResults.Books)
                        resultsListBox.Items.Add(book);

                    int returnCount = searchResults.Books.Count;
                    int listCount = searchResults.StartIndex + returnCount;
                    int totalCount = searchResults.TotalCount;

                    if (listCount < totalCount)
                    {
                        statusTextBlock.Text = String.Format("Showing first {0} matches out of {1}", listCount, totalCount);

                        nextSearchButton.Visibility = Visibility.Visible;
                        nextSearchButton.IsEnabled = true;
                        searchParams.Index = listCount;
                        searchParams.SearchKey = searchResults.SearchKey;

                        if (returnCount + listCount < totalCount)
                            nextSearchButton.Content = String.Format("get next {0} {1}", returnCount, MatchesText(returnCount));
                        else
                        {
                            int lastCount = totalCount - listCount;
                            nextSearchButton.Content = String.Format("get last {0} {1}", lastCount, MatchesText(lastCount));
                        }
                    }
                    else
                    {
                        statusTextBlock.Text = String.Format("Showing {0} {1}", listCount, MatchesText(listCount));
                        nextSearchButton.Visibility = Visibility.Collapsed;
                    }
                }

                statusTextBlock.Text += String.Format(" ({0} seconds)", (searchResults.FinishTime - searchResults.StartTime).TotalSeconds.ToString("F1"));
            }
        }

        string MatchesText(int num)
        {
            return num == 1 ? "match" : "matches";
        }

        void OnNextSearchButtonClick(object sender, EventArgs args)
        {
            EnableForSearch(false);
            nextSearchButton.IsEnabled = false;

            proxy.SearchAsync(searchParams.Title, searchParams.Author, searchParams.Index, searchParams.SearchKey);
        }

        void EnableForSearch(bool enable)
        {
            searchButton.IsEnabled = enable;
            titleTextBox.IsEnabled = enable;
            authorTextBox.IsEnabled = enable;
            progressBar.Visibility = enable ? Visibility.Collapsed : Visibility.Visible;
            progressBar.IsIndeterminate = !enable;
        }

        void OnResultsListBoxSelectionChanged(object sender, SelectionChangedEventArgs args)
        {
            if (resultsListBox.SelectedItem != null)
            {
                GutenbergBook book = resultsListBox.SelectedItem as GutenbergBook;
                resultsListBox.SelectedItem = null;

                StateDictionaryHelper.SafeAdd(PhoneApplicationService.Current.State, "downloadBook", book);
                (Application.Current.RootVisual as Frame).Navigate(new Uri("/DownloadBookPage.xaml", UriKind.Relative));
            }          
        }
    }
}
